#include "IOTCDevice.h"
#include "P2PTunnelServer.h"
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#define ENABLE_TUNNEL_SERVICE 0

#define MAX_CONNECTION 20
#define IOTC_AUTH_KEY "00000000"
#define TUNNEL_USERNAME "admin"
#define TUNNEL_PASSWORD "testpwd"

#define HTTP_SERVICE_PORT 80
#define HTTPS_SERVICE_PORT 443
#define SSH_SERVICE_PORT 22
#define TELNET_SERVICE_PORT 23

//########################################################
//# IOTC Device
//########################################################
static int iotcDeviceCreate(const char *uid) {
    int ret = 0;
    DeviceLoginInput optionalInput;

    optionalInput.cb = sizeof(DeviceLoginInput);
    optionalInput.authentication_type = AUTHENTICATE_BY_KEY;
    memcpy(optionalInput.auth_key, IOTC_AUTH_KEY, IOTC_AUTH_KEY_LENGTH);

    while (1) {
        ret = IOTC_Device_LoginEx(uid, &optionalInput);
        printf("IOTC_Device_LoginEx() ret = %d\n", ret);

        if (ret == IOTC_ER_NoERROR)
            break;
        else
            Sleep(3000);
    }

    return 0;
}

//########################################################
//# P2PTunnel Server
//########################################################
static bool gSidUsed[MAX_CONNECTION] = {0};

static void WINAPI tunnelServerAuthentication(const char *cszAccount, char *cszPassword, uint32_t nPasswordMaxLength, const void *pArg) {
    printf("[%s] cszAccount=%s, nPasswordMaxLength=%d, pArg=%s\n", __func__, cszAccount, nPasswordMaxLength, (char *)pArg);
    if (strcmp(cszAccount, TUNNEL_USERNAME) == 0) {
        strcpy(cszPassword, TUNNEL_PASSWORD);
    } else {
        printf("unknow account, reject connect\n");
        // do nothing to reject connect
    }
}

static void WINAPI tunnelStatusHandle(int nErrorCode, int nSID, void *pArg) {
    if (nErrorCode == TUNNEL_ER_DISCONNECTED) {
        printf("SID[%d] TUNNEL_ER_DISCONNECTED Log file here!\n", nSID);
        P2PTunnelServer_Disconnect(nSID);
        IOTC_Session_Close(nSID); // if AV module working on this session, AV will be disconnected.
        gSidUsed[nSID] = false;
        if (pArg != NULL)
            printf("pArg = %s\n", (char *)pArg);
    }
}

static void WINAPI tunnelSessionInfoExHandle(sP2PTunnelSessionInfoEx *sSessionInfo, const void *pArg) {
    int ret = 0;
    printf("[%s]\n", __func__);
    printf("  Connection Mode = %d, NAT type = %d\n", sSessionInfo->nMode, sSessionInfo->nNatType);
    printf("  P2PTunnel Version = %X, SID = %d\n", (unsigned int)sSessionInfo->nVersion, sSessionInfo->nSID);
    printf("  IP Address = %s:%d\n", sSessionInfo->szRemoteIP, sSessionInfo->nRemotePort);
    if (pArg != NULL) {
        printf("  pArg = %s\n", (char *)pArg);
    }

    gSidUsed[sSessionInfo->nSID] = true;
    ret = P2PTunnel_SetBufSize(sSessionInfo->nSID, 5120000);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnel_SetBufSize() SID[%d] error[%d]\n", sSessionInfo->nSID, ret);
    }

    return;
}

static int tunnelServerManualListen(void) {

    int ret = 0;
    int nSID = IOTC_Listen(5000);
    if (nSID > -1) {
        ret = P2PTunnelServer_Listen(nSID, 1);
        printf("P2PTunnelServer_Listen() ret = %d\n", ret);
    } else if (nSID == IOTC_ER_EXCEED_MAX_SESSION) {
        printf("IOTC_Listen() ret = IOTC_ER_EXCEED_MAX_SESSION\n");
    } else if (nSID == IOTC_ER_TIMEOUT) {
        printf("IOTC_Listen() ret = IOTC_ER_TIMEOUT\n");
    } else {
        printf("IOTC_Listen() error[%d]\n", nSID);
        return -1;
    }

    if (ret != TUNNEL_ER_NoERROR) {
        IOTC_Session_Close(nSID); // if AV module working on this session, AV will be disconnected.
    }
    return ret;
}

int WINAPI TunnelPortVerifyCB(uint16_t nServicePort, const void *pArg) {
    int ret = 0;

    switch (nServicePort) {
    case HTTP_SERVICE_PORT:
    case SSH_SERVICE_PORT:
    case TELNET_SERVICE_PORT:
        printf("[%s] port[%d] pass\n", __func__, nServicePort);
        break;
    default:
        printf("[%s] port[%d] reject\n", __func__, nServicePort);
        ret = -1;
        break;
    }
    return ret;
}

#if ENABLE_TUNNEL_SERVICE
uint16_t WINAPI TunnelServiceVerifyCB(const char *cszServiceID, const void *pArg) {
    if (strcmp(cszServiceID, "https") == 0) {
        printf("[%s] service[%s] pass, service_port[%u]\n", __func__, cszServiceID, HTTPS_SERVICE_PORT);
        return HTTPS_SERVICE_PORT;
    }
    return 0;
}
#endif

//########################################################
//# Main
//########################################################
int main(int argc, char *argv[]) {
    int ret = 0;
    char *uid = NULL;
    char *license_key = NULL;
    LogAttr log_attr = {0};
    char log_path[] = "./device_log.txt";
    char *args = strdup(__FILE__);

    if (argc != 3) {
        printf("%s <LICENSE_KEY> <UID>\n", argv[0]);
        return -1;
    }

    license_key = argv[1];
    uid = argv[2];
    if (strlen(uid) != 20) {
        printf("invalid UID\n");
        free(args);
        return -1;
    }

    //===== set license key for enable TUTK SDK =====
    ret = TUTK_SDK_Set_License_Key(license_key);
    printf("TUTK_SDK_Set_License_Key() ret[%d]\n", ret);
    if (ret < 0) {
        free(args);
        return -1;
    }

    //===== create IOTC device =====
    IOTC_Set_Max_Session_Number(MAX_CONNECTION);

    ret = IOTC_Initialize2(0);
    if (ret != IOTC_ER_NoERROR) {
        printf("IOTC_Initialize2() error[%d]!\n", ret);
        return -1;
    }

    ret = iotcDeviceCreate(uid);
    if (ret < 0) {
        IOTC_DeInitialize();
        free(args);
        return -1;
    }

    //==== create P2PTunnel server with manual mode =====
    log_attr.path = log_path;
    ret = P2PTunnel_Set_Log_Attr(log_attr);
    printf("P2PTunnel_Set_Log_Attr() ret[%d]\n", ret);

    ret = P2PTunnelSetConnectionOption(TUNNEL_CONNECT_MANUAL);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelSetConnectionOption() ret[%d]\n", ret);
        IOTC_DeInitialize();
        free(args);
        return -1;
    }

    ret = P2PTunnelServerInitialize2(MAX_CONNECTION, 0);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServerInitialize2() ret[%d]\n", ret);
        IOTC_DeInitialize();
        free(args);
        return -1;
    }

    ret = P2PTunnelServer_Register_Port_Verify(TunnelPortVerifyCB, NULL);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServer_Register_Port_Verify() error[%d]!\n", ret);
        return -1;
    }

#if ENABLE_TUNNEL_SERVICE
    ret = P2PTunnelServer_Register_Service_Verify(TunnelServiceVerifyCB, NULL);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServer_Register_Service_Verify() error[%d]!\n", ret);
        return -1;
    }
#endif

    P2PTunnelServer_GetStatus(tunnelStatusHandle, args);

    ret = P2PTunnelServer_Start_Ex(uid, tunnelServerAuthentication, tunnelSessionInfoExHandle, args);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServer_Start_Ex() error[%d]!\n", ret);
        P2PTunnelServerDeInitialize();
        IOTC_DeInitialize();
        free(args);
        return -1;
    } else {
        printf("P2PTunnelServer_Start_Ex() Success, I can connected by Internet.\n");
    }

    while (1) {
        ret = tunnelServerManualListen();
        for (int i = 0; i < MAX_CONNECTION; i++) {
            if (gSidUsed[i]) {
                int access_time = P2PTunnel_LastIOTime(i);
                if (access_time < 0) {
                    printf("P2PTunnel_LastIOTime() error[%d]\n", access_time);
                    gSidUsed[i] = false;
                } else {
                    printf("SID %d:%u, ", i, access_time);
                }
            }
        }
    }
    
    P2PTunnelServer_Stop();
    P2PTunnelServerDeInitialize();
    IOTC_DeInitialize();
    free(args);
    return 0;
}
